module data_mod
use sep
use sep3d_struct_mod
use sinc_f90_mod
use mpi_sep
implicit none


type data_par_type
  real :: dhx,dhy,dw,dsx,dsy
  real :: ohx,ohy,ow,osx,osy
  integer :: nhx,nhy,nw,nsx,nsy
  integer ::MPI_size
end type data_par_type

type plane_par_type
  real ::plane_os_x,plane_ds_x,plane_os_y,plane_ds_y
  integer :: plane_ns_x,plane_ns_y
  integer :: MPI_size
end type plane_par_type

logical :: migration
type(sep3d),private :: data,wavlt
type(data_par_type),private :: data_par
type(plane_par_type),private:: plane_par
complex,allocatable,private :: wavelet(:)

contains



subroutine data_mod_init(ns,nw,dx,dy,dw,w0,i_node)
integer ::ns,nw,i_node
real :: dx,dy,dw,w0

integer :: size_real,size_logical,size_int
integer,external :: auxin,sreed
integer :: ierr
real :: wlt_o, wlt_d
integer :: wlt_n

call MPI_TYPE_SIZE(MPI_REAL,size_real,ierr)
call MPI_TYPE_SIZE(MPI_LOGICAL,size_logical,ierr)
call MPI_TYPE_SIZE(MPI_INTEGER,size_int,ierr)
data_par%MPI_size=10*size_real+6*size_int
plane_par%MPI_size=4*size_real+3*size_int


if (i_node==0) then
    if (-1==auxin("Data") ) call seperr("Must provide Data file")
    call init_sep3d("Data",data,"INPUT")
    ! data : hx,hy,w,sx,sy
    data_par%dhx=data%d(1);data_par%ohx=data%o(1); data_par%nhx=data%n(1)
    data_par%dhy=data%d(2);data_par%ohy=data%o(2); data_par%nhy=data%n(2)
    data_par%dw =data%d(3);data_par%ow =data%o(3); data_par%nw =data%n(3)
    data_par%dsx=data%d(4);data_par%osx=data%o(4); data_par%nsx=data%n(4)
    if (size(data%n)==4 ) then
      data_par%dsy=100;data_par%osy=0; data_par%nsy=1
    else
      data_par%dsy=data%d(5);data_par%osy=data%o(5); data_par%nsy=data%n(5)
    end if
end if

call MPI_BCAST(data_par,data_par%MPI_Size,MPI_BYTE,0,MPI_COMM_WORLD,ierr)
ns=data_par%nsx*data_par%nsy
dx=data_par%dhx;  dy=data_par%dhy
dw=data_par%dw;   w0=data_par%ow; nw=data_par%nw

if (i_node==0 ) then
  call data_read_wavelet(wlt_o,wlt_d,wlt_n)
  if ( wlt_n /= nw .or. wlt_d /= dw .or. wlt_o /=w0) call seperr("Wavelet should have same frequency as Data file")
else
  allocate(wavelet( data_par%nw) )
end if
call MPI_BCAST(wavelet,data_par%nw,MPI_COMPLEX,0,MPI_COMM_WORLD,ierr)
end subroutine data_mod_init


subroutine data_mod_init_modeling(nw,dx,dw,w0)
integer :: nw
real :: sx,dx,dw,w0

real :: wlt_o, wlt_d
integer :: wlt_n

call data_read_wavelet(wlt_o,wlt_d,wlt_n)
w0=wlt_o; nw=wlt_n; dw=wlt_d
data_par%dhx=dx; data_par%dw=dw; data_par%ow=w0; data_par%nw=nw
end subroutine data_mod_init_modeling


subroutine data_source_modeling(sx,min_x,nx,wld,iw)
integer :: nx,iw
real :: sx,min_x
complex :: wld(:,:)

integer ix,iy

iy=1
ix=(sx-min_x)/data_par%dhx+1
!write(*,*) "source_ix",ix,"y",iy
if (ix <1 .or. ix>nx ) call seperr("err create data_source")
!wait
wld=0.0
wld(ix,iy)=wavelet(iw)
if (ix+1 <=nx) wld(ix+1,iy)=0.5*wavelet(iw)
if (ix-1 >0  ) wld(ix-1,iy)=0.5*wavelet(iw)
end subroutine data_source_modeling


subroutine data_read_wavelet(wlt_o,wlt_d,wlt_n)
real :: wlt_o,wlt_d
integer :: wlt_n
integer,external :: auxin,sreed

if (-1==auxin("Wavelet") ) call seperr("Must provide Wavelet file")
call init_sep3d("Wavelet",wavlt,"INPUT")
wlt_o=wavlt%o(1); wlt_d=wavlt%d(1); wlt_n=wavlt%n(1)
allocate(wavelet( wavlt%n(1) ))
if (sreed("Wavelet",wavelet,wavlt%n(1)*8 )/= wavlt%n(1)*8 ) call seperr("err read Wavelet")
end subroutine data_read_wavelet



subroutine data_1shot_init(i_s,mig_x_min,mig_y_min,nx,ny)
integer:: i_s
integer:: nx,ny
real   :: mig_x_min,mig_y_min

integer :: i_sx,i_sy
real    :: sx,sy

i_sx=mod(i_s-1,data_par%nsx)+1;                i_sy=(i_s-1)/data_par%nsx+1
!write(*,*) "i_sx",i_sx,"i_sy",i_sy
sx=(i_sx-1)*data_par%dsx+data_par%osx;     sy=(i_sy-1)*data_par%dsy+data_par%osy
mig_x_min=sx+data_par%ohx;                 mig_y_min=sy+data_par%ohy
!mig_x_min=data_par%ohx;                    mig_y_min=data_par%ohy
nx=data_par%nhx;                           ny=data_par%nhy
end subroutine data_1shot_init


subroutine plane_1shot_init(i_s,mig_x_min,mig_y_min,nx,ny,plane_center,rotate_angle,i_node)
integer:: i_s,i_node
integer:: nx,ny
real   :: mig_x_min,mig_y_min,plane_center,rotate_angle
integer :: ierr

integer :: i_px,i_py
real :: px,py,sin_rotate_angle

mig_x_min=data_par%ohx;                 mig_y_min=data_par%ohy
nx=data_par%nhx;                           ny=data_par%nhy
i_px=mod(i_s-1,data_par%nsx)+1;                i_py=(i_s-1)/data_par%nsx+1
px=(i_px-1)*data_par%dsx+data_par%osx;     py=(i_py-1)*data_par%dsy+data_par%osy


sin_rotate_angle=px*1500.0
rotate_angle=asin(sin_rotate_angle)
if ( abs(rotate_angle) <10.0/180.0*3.1415926) then
     rotate_angle=0.0
end if
!if (migration) then
if (i_node==0) then
  call from_aux("Data", "plane_os_x", plane_par%plane_os_x, 0.)
  call from_aux("Data", "plane_ds_x", plane_par%plane_ds_x, 1.)
  call from_aux("Data", "plane_ns_x", plane_par%plane_ns_x, 1)
  call from_aux("Data", "plane_os_y", plane_par%plane_os_y, 0.)
  call from_aux("Data", "plane_ds_y", plane_par%plane_ds_y, 1.)
  call from_aux("Data", "plane_ns_y", plane_par%plane_ns_y, 1)
end if
!else
!  if (i_node==0) then
!    call from_param("plane_os_x", plane_par%plane_os_x, data_par%ohx)
!    call from_param("plane_ds_x", plane_par%plane_ds_x, data_par%dhx)
!    call from_param("plane_ns_x", plane_par%plane_ns_x, data_par%nhx)
!    call from_param("plane_os_y", plane_par%plane_os_y, 0.)
!    call from_param("plane_ds_y", plane_par%plane_ds_y, 1.)
!    call from_param("plane_ns_y", plane_par%plane_ns_y, 1)
!  end if
!end if
call MPI_BCAST(plane_par,plane_par%MPI_Size,MPI_BYTE,0,MPI_COMM_WORLD,ierr)
plane_center=plane_par%plane_os_x
write(*,*) "plane_os,x",plane_par%plane_os_x, plane_par%plane_ds_x,plane_par%plane_ns_x
end subroutine plane_1shot_init

!==== read point source ===================
subroutine data_source(min_x,min_y,nx,ny,iw,i_s,wld)
real :: min_x,min_y
integer :: nx,ny,iw,i_s
complex :: wld(:,:)

integer :: i_sx,i_sy
real    :: sx,sy
integer :: ix,iy

i_sx=mod(i_s-1,data_par%nsx)+1;                       i_sy=(i_s-1)/data_par%nsx+1
sx=(i_sx-1)*data_par%dsx+data_par%osx;               sy=(i_sy-1)*data_par%dsy+data_par%osy
!write(*,*) "min_x",min_x,"sx",sx
!write(*,*) "source sx",sx,"source sy",sy
ix=(sx-min_x)/data_par%dhx+1; iy=(sy-min_y)/data_par%dhy+1
!write(*,*) "source_ix",ix,"y",iy
if (ix <1 .or. iy <1 .or. ix>nx .or. iy>ny) call seperr("err create data_source")
!wait
wld=0.0
wld(ix,iy)=wavelet(iw)
if (ix+1 <=nx) wld(ix+1,iy)=0.5*wavelet(iw)
if (ix-1 >0  ) wld(ix-1,iy)=0.5*wavelet(iw)
end subroutine data_source
!==== read point source ====================

!==== read plane source ====================
subroutine plane_source(min_x,min_y,nx,ny,iw,i_p,wld,center)
real :: min_x,min_y,center
integer :: nx,ny,iw,i_p
complex :: wld(:,:)

real :: x_dt,y_dt,t_x
integer :: iy,ix,isx,isy,i_px,i_py
real :: plane_s_x
real :: px,py
real :: w

w=(iw-1)*data_par%dw+data_par%ow
i_px=mod(i_p-1,data_par%nsx)+1;                i_py=(i_p-1)/data_par%nsx+1
!write(*,*) "i_px",i_px,"i_py",i_py
px=(i_px-1)*data_par%dsx+data_par%osx;     py=(i_py-1)*data_par%dsy+data_par%osy

!plane_os_x=data_par%osx; plane_ds_x=data_par%dsx; plane_ns_x=data_par%nsx
!plane_os_y=data_par%osy; plane_ds_y=data_par%dsy; plane_ns_y=data_par%nsy
!write(*,*) "px",px
wld=0.0
x_dt=plane_par%plane_ds_x*px
y_dt=0.0
!write(*,*)"x_dt",x_dt
!write(*,*) "plane_os_x", plane_os_x, "plane_ds_x", plane_ds_x,"plane_ns_x", plane_ns_x
do isy=1,plane_par%plane_ns_y
  iy=isy
  do isx=1,plane_par%plane_ns_x
      plane_s_x=(isx-1)*plane_par%plane_ds_x+plane_par%plane_os_x
      ix=(plane_s_x-min_x)/abs(data_par%dhx)+1
      t_x=(isx-1)*x_dt
      !t_x=(plane_s_x-center)*px
      wld(ix,iy)=wld(ix,iy)+wavelet(iw)*cmplx(cos(w*t_x),sin(w*t_x))
  end do
end do
end subroutine plane_source

!==== read plane source ====================


subroutine data_receiver(min_x,min_y,nx,ny,iw,i_s,wld)

real :: min_x,min_y
integer :: nx,ny,iw,i_s
complex :: wld(:,:)

complex,allocatable:: tmp_wld(:,:)
integer :: i_sx,i_sy
real    :: sx,sy,max_x,max_y,min_x_d,min_y_d,max_x_d,max_y_d
real    :: max_x_read,max_y_read,min_x_read,min_y_read
integer :: nx_read,ny_read,i_ox_read,i_oy_read,i_ox_wld,i_oy_wld
integer :: n_block,blocksize
integer,external:: sseek_block,sreed

allocate( tmp_wld(data_par%nhx,data_par%nhy ) )
wld=0.0
i_sx=mod(i_s-1,data_par%nsx )+1;                       i_sy=(i_s-1)/data_par%nsx+1
sx=(i_sx-1)*data_par%dsx+data_par%osx;               sy=(i_sy-1)*data_par%dsy+data_par%osy
!write(*,*)"is",i_s,"sx=",sx,"sy=",sy
max_x=min_x+(nx-1)*data_par%dhx;                  max_y=min_y+(ny-1)*data_par%dhy
min_x_d=sx+data_par%ohx;                          min_y_d=sy+data_par%ohy
!min_x_d=data_par%ohx; min_y_d=data_par%ohy
max_x_d=min_x_d+data_par%dhx*(data_par%nhx-1);       max_y_d=min_y_d+data_par%dhy*(data_par%nhy-1)
max_x_read=min(max_x_d,max_x);                 max_y_read=min(max_y_d,max_y)
min_x_read=max(min_x_d,min_x);                 min_y_read=max(min_y_d,min_y)
nx_read=(max_x_read-min_x_read)/data_par%dhx+1;   ny_read=(max_y_read-min_y_read)/data_par%dhy+1
i_ox_read=(min_x_read-min_x_d)/data_par%dhx+1;    i_oy_read=(min_y_read-min_y_d)/data_par%dhy+1
i_ox_wld=(min_x_read-min_x)/data_par%dhx+1;       i_oy_wld=(min_y_read-min_y)/data_par%dhy+1
!write(*,*) "i_oxwld",i_ox_wld,"y",i_oy_wld,"nx",nx_read,"ny",ny_read,"oxr",i_ox_read,"oyr",i_oy_read
n_block=(i_s-1)*data_par%nw+iw-1;                       blocksize=data_par%nhy*data_par%nhx*8
!write(*,*)"n_block",n_block,blocksize
if ( sseek_block("Data",n_block,blocksize,0) /=n_block) call seperr("data seek wrong")
if ( sreed("Data",tmp_wld(:,:),blocksize )/= blocksize) call seperr("data read wrong")
wld(i_ox_wld:i_ox_wld+nx_read-1,i_oy_wld:i_oy_wld+ny_read-1)= &
 tmp_wld(i_ox_read:i_ox_read+nx_read-1,i_oy_read:i_oy_read+ny_read-1)

deallocate(tmp_wld)
end subroutine data_receiver


subroutine plane_receiver(min_x,min_y,nx,ny,iw,i_s,wld)
real :: min_x,min_y
integer :: nx,ny,iw,i_s
complex :: wld(:,:)

integer :: i_px,i_py
real    :: min_x_d,min_y_d,max_x_d,max_y_d,max_x,max_y
real    :: max_x_read,max_y_read,min_x_read,min_y_read
integer :: nx_read,ny_read,i_ox_read,i_oy_read,i_ox_wld,i_oy_wld
integer :: n_block,blocksize
integer,external:: sseek_block,sreed

complex,allocatable :: tmp_wld(:,:)
allocate(tmp_wld(data_par%nhx,data_par%nhy))
max_x=min_x+(nx-1)*abs(data_par%dhx);                     max_y=min_y+(ny-1)*abs(data_par%dhy)
i_px=mod(i_s-1,data_par%nsx )+1;                     i_py=(i_s-1)/data_par%nsx+1
min_x_d=data_par%ohx;                                min_y_d=data_par%ohy
max_x_d=min_x_d+abs(data_par%dhx)*(data_par%nhx-1);       max_y_d=min_y_d+abs(data_par%dhy)*(data_par%nhy-1)
max_x_read=min(max_x_d,max_x);                 max_y_read=min(max_y_d,max_y)
min_x_read=max(min_x_d,min_x);                 min_y_read=max(min_y_d,min_y)
nx_read=(max_x_read-min_x_read)/abs(data_par%dhx)+1;   ny_read=(max_y_read-min_y_read)/abs(data_par%dhy)+1
i_ox_read=(min_x_read-min_x_d)/abs(data_par%dhx)+1;    i_oy_read=(min_y_read-min_y_d)/abs(data_par%dhy)+1
i_ox_wld=(min_x_read-min_x)/abs(data_par%dhx)+1;       i_oy_wld=(min_y_read-min_y)/abs(data_par%dhy)+1
n_block=(i_s-1)*data_par%nw+iw-1;        blocksize=data_par%nhy*data_par%nhx*8
if ( sseek_block("Data",n_block,blocksize,0) /=n_block) call seperr("data seek wrong")
if ( sreed("Data",tmp_wld(:,:),blocksize )/= blocksize) call seperr("data read wrong")
wld(i_ox_wld:i_ox_wld+nx_read-1,i_oy_wld:i_oy_wld+ny_read-1)= &
 tmp_wld(i_ox_read:i_ox_read+nx_read-1,i_oy_read:i_oy_read+ny_read-1)
deallocate(tmp_wld)
end subroutine plane_receiver



subroutine rotate(z0_s,z0_s_rotate,rotate_angle,ix_rotate_center,dx,dz,nx,nx_rotate,nz_rotate)
complex :: z0_s(:,:), z0_s_rotate(:,:,:)
integer :: nx,nx_rotate,nz_rotate,ix_rotate_center
real    :: dx,dz,rotate_angle

integer ::ix,ix_rotate,iz_rotate,i_ix_rotate,i_iz_rotate
real    ::x,x_rotate,z_rotate,wx_rotate,wz_rotate
real    :: sincx(16),sincz(16)
complex,allocatable :: z0_s_smooth(:,:)
integer :: nd
real :: ddx

nd=16

ddx=dx/nd
allocate( z0_s_smooth(nd*nx,1))

z0_s_smooth=0.0
do ix=1,nd*nx
  wx_rotate=mod(ix-1,nd)/nd
  ix_rotate=(ix-1)/nd+1
  call sinc_f90_mksinc(wx_rotate,16,sincx)
  do i_ix_rotate=1,16
     if (ix_rotate-8+i_ix_rotate >0 .and. ix_rotate-8+i_ix_rotate <nx+1 ) then
        z0_s_smooth(ix,1)=z0_s_smooth(ix,1)+z0_s(ix_rotate-8+i_ix_rotate,1)*sincx(i_ix_rotate)
     end if
  end do
end do


z0_s_rotate=0.0

do ix=1,nx*nd
  if (rotate_angle >=0 ) then
    x=(ix-1)*ddx
  else 
    x=(nx*nd-ix)*ddx
  end if
  x_rotate=x*cos(abs(rotate_angle)); z_rotate=x*sin(abs(rotate_angle))
  ix_rotate=(x_rotate-0.0)/dx+1;iz_rotate=(z_rotate-0.0)/dz+1
  wx_rotate=(x_rotate-(ix_rotate-1)*dx)/dx
  wz_rotate=(z_rotate-(iz_rotate-1)*dz)/dz
  call sinc_f90_mksinc(wx_rotate,16,sincx); call sinc_f90_mksinc(wz_rotate,16,sincz)
  do i_ix_rotate=1,16
    if (ix_rotate-8+i_ix_rotate+ix_rotate_center >0 .and. ix_rotate-8+i_ix_rotate+ix_rotate_center <nx_rotate+1) then
      do i_iz_rotate=1,16
         if (iz_rotate-8+i_iz_rotate >0 .and. iz_rotate-8+i_iz_rotate <nz_rotate+1) then
           z0_s_rotate(ix_rotate-8+i_ix_rotate+ix_rotate_center,1,iz_rotate-8+i_iz_rotate)= & 
           z0_s_rotate(ix_rotate-8+i_ix_rotate+ix_rotate_center,1,iz_rotate-8+i_iz_rotate)+ &
           z0_s_smooth(ix,1)*sincx(i_ix_rotate)*sincz(i_iz_rotate)
         end if
      end do
    end if
  end do
!  do i_iz_rotate=1,16
!    if (iz_rotate-8+i_iz_rotate >0 .and. iz_rotate-8+i_iz_rotate <nz_rotate+1) then
!       z0_s_rotate(ix_rotate_center,1,iz_rotate-8+i_iz_rotate)= &
!       z0_s_rotate(ix_rotate_center,1,iz_rotate-8+i_iz_rotate)+ &
!       z0_s_smooth(ix,1)*sincz(i_iz_rotate)
!    end if
!  end do
end do



!z0_s_rotate=0.0
!do ix=1,nx
!  x=(ix-1)*dx
!  x_rotate=x*cos(rotate_angle); z_rotate=x*sin(rotate_angle)
!  ix_rotate=(x_rotate-0.0)/dx+1;iz_rotate=(z_rotate-0.0)/dz+1
!  wx_rotate=(x_rotate-(ix_rotate-1)*dx)/dx
!  wz_rotate=(z_rotate-(iz_rotate-1)*dz)/dz
!  call sinc_f90_mksinc(wx_rotate,8,sincx); call sinc_f90_mksinc(wz_rotate,8,sincz)
!  do i_ix_rotate=1,8
!    if (ix_rotate-4+i_ix_rotate >0 .and. ix_rotate-4+i_ix_rotate <nx_rotate+1) then
!      do i_iz_rotate=1,8
!         if (iz_rotate-4+i_iz_rotate >0 .and. iz_rotate-4+i_iz_rotate <nz_rotate+1) then
!           z0_s_rotate(ix_rotate-4+i_ix_rotate+ix_rotate_center,1,iz_rotate-4+i_iz_rotate)= & 
!           z0_s_rotate(ix_rotate-4+i_ix_rotate+ix_rotate_center,1,iz_rotate-4+i_iz_rotate)+ &
!           z0_s(ix,1)*sincx(i_ix_rotate)*sincz(i_iz_rotate)
!         end if
!      end do
!    end if
!  end do
!end do


deallocate(z0_s_smooth)
end subroutine rotate



subroutine img_rotate(image,image_rotate,rotate_angle,ix_rotate_center,dx,dz,nx,nz,nx_rotate,nz_rotate)
real:: image(:,:,:,:,:),image_rotate(:,:,:,:,:)
real :: rotate_angle,dx,dz
integer :: ix_rotate_center, nx_rotate,nz_rotate,nx,nz

integer ::ix,iz,ix_rotate,iz_rotate,i_ix_rotate,i_iz_rotate,ii_ix_rotate,ii_iz_rotate
real    ::x,z,x_rotate,z_rotate,wx_rotate,wz_rotate
real    :: sincx(16),sincz(16)



image_rotate=0.0

do iz=1,nz
  do ix=1,nx
    if (rotate_angle >=0 ) then
      x=(ix-1)*dx 
    else
      x=(nx-ix)*dx
    end if
    z=(iz-1)*dz
    x_rotate=x*cos(abs(rotate_angle))-z*sin(abs(rotate_angle))
    z_rotate=x*sin(abs(rotate_angle))+z*cos(abs(rotate_angle))
    ix_rotate=(x_rotate-0.0)/dx+1;                    iz_rotate=(z_rotate-0.0)/dz+1
    wx_rotate=(x_rotate-(ix_rotate-1)*dx)/dx;         wz_rotate=(z_rotate-(iz_rotate-1)*dz)/dz  
    call sinc_f90_mksinc(wx_rotate,16,sincx); call sinc_f90_mksinc(wz_rotate,16,sincz)
    do ii_ix_rotate=1,16
       i_ix_rotate=ix_rotate-8+ii_ix_rotate+ix_rotate_center
      if (i_ix_rotate >0 .and. i_ix_rotate <nx_rotate+1) then
        do ii_iz_rotate=1,16
          i_iz_rotate= iz_rotate-8+ii_iz_rotate
          if (i_iz_rotate >0 .and. i_iz_rotate <nz_rotate+1) then
            image_rotate(iz,:,:,ix,1)=image_rotate(iz,:,:,ix,1)+&
            image(:,:,i_ix_rotate,1,i_iz_rotate)*sincx(ii_ix_rotate)*sincz(ii_iz_rotate)
          end if
        end do
      end if
    end do  
  end do
end do

!do iz=1,nz
!  do ix=1,nx
!    x=(ix-1)*dx; z=(iz-1)*dz
!    x_rotate=x*cos(rotate_angle)-z*sin(rotate_angle); z_rotate=x*sin(rotate_angle)+z*cos(rotate_angle)
!    ix_rotate=(x_rotate-0.0)/dx+1;                    iz_rotate=(z_rotate-0.0)/dz+1
!    wx_rotate=(x_rotate-(ix_rotate-1)*dx)/dx;         wz_rotate=(z_rotate-(iz_rotate-1)*dz)/dz  
!    call sinc_f90_mksinc(wx_rotate,8,sincx); call sinc_f90_mksinc(wz_rotate,8,sincz)
!    do ii_ix_rotate=1,8
!       i_ix_rotate=ix_rotate-4+ii_ix_rotate+ix_rotate_center
!      if (i_ix_rotate >0 .and. i_ix_rotate <nx_rotate+1) then
!        do ii_iz_rotate=1,8
!          i_iz_rotate= iz_rotate-4+ii_iz_rotate
!          if (i_iz_rotate >0 .and. i_iz_rotate <nz_rotate+1) then
!            image_rotate(iz,:,:,ix,1)=image_rotate(iz,:,:,ix,1)+&
!            image(:,:,i_ix_rotate,1,i_iz_rotate)*sincx(ii_ix_rotate)*sincz(ii_iz_rotate)
!          end if
!        end do
!      end if
!    end do  
!  end do
!end do


end subroutine img_rotate

subroutine shot2plane_init(ns,nw,dx,dy,dw,w0,i_node)
integer ::ns,nw,i_node
real :: dx,dy,dw,w0

real :: wlt_o, wlt_d
integer :: wlt_n
integer :: size_real,size_logical,size_int
integer,external :: auxin,sreed
integer :: ierr

!call MPI_TYPE_SIZE(MPI_REAL,size_real,ierr)
!call MPI_TYPE_SIZE(MPI_LOGICAL,size_logical,ierr)
!call MPI_TYPE_SIZE(MPI_INTEGER,size_int,ierr)
!data_par%MPI_size=10*size_real+6*size_int
!plane_par%MPI_size=4*size_real+3*size_int

if (i_node==0) then
  call init_sep3d("Data",data,"INPUT")
  data_par%dw =data%d(3); data_par%ow =data%o(3); data_par%nw =data%n(3)
  data_par%dhx=data%d(1); data_par%ohx=data%o(1); data_par%nhx=data%n(1)
  data_par%dhy=data%d(2); data_par%ohy=data%o(2); data_par%nhy=data%n(2)
  data_par%dsx=data%d(4); data_par%osx=data%o(4); data_par%nsx=data%n(4)
  data_par%dsy=data%d(5); data_par%osy=data%o(5); data_par%nsy=data%n(5)
  data_par%dsy=10.0;      data_par%osy=0.0;   data_par%nsy=1
end if  
!call MPI_BCAST(data_par,data_par%MPI_Size,MPI_BYTE,0,MPI_COMM_WORLD,ierr)
ns=data_par%nsx*data_par%nsy
dx=data_par%dhx;  dy=data_par%dhy
dw=data_par%dw;   w0=data_par%ow; nw=data_par%nw

end subroutine shot2plane_init


subroutine plane_write()
type(sep3d):: plane_data
real :: opx,dpx,opy,dpy
integer :: npx,npy
real :: x_min,y_min,x_max,y_max,tmp_x_min,tmp_y_min
integer :: plane_nx,plane_ny,tmp_plane_nx,tmp_plane_ny
complex,allocatable :: wld(:,:,:)
integer :: ipx,ipy
real    :: px,py
real ::plane_osx,plane_dsx,plane_osy,plane_dsy
integer ::plane_nsx,plane_nsy
integer :: blocksize
integer,external :: srite

call from_param('opx',opx,0.0)
call from_param('dpx',dpx,1.0)
call from_param('npx',npx,30)
call from_param('opy',opy,0.0)
call from_param('dpy',dpy,1.0)
call from_param('npy',npy,1)


plane_osx=data_par%osx; plane_dsx=data_par%dsx; plane_nsx=data_par%nsx
plane_osy=data_par%osy; plane_dsy=data_par%dsy; plane_nsy=data_par%nsy
tmp_x_min=min(data_par%osx,data_par%osx+data_par%dsx*(data_par%nsx-1))&
+min(data_par%ohx+data_par%dhx*(data_par%nhx-1),data_par%ohx)
!tmp_x_min=data_par%ohx
write(*,*) "tmp_x_min",tmp_x_min
tmp_y_min=min(data_par%osy+data_par%dsy*(data_par%nsy-1),data_par%osy)&
+min(data_par%ohy+data_par%dhy*(data_par%nhy-1),data_par%ohy)
!tmp_y_min=data_par%ohy
call from_param('plane_x_min',x_min,tmp_x_min)
call from_param('plane_y_min',y_min,tmp_y_min)
x_max=max(data_par%osx,data_par%osx+(data_par%nsx-1)*data_par%dsx)&
+max(data_par%ohx,data_par%ohx+(data_par%nhx-1)*data_par%dhx)
!x_max=data_par%ohx+(data_par%nhx-1)*data_par%dhx
write(*,*) "x_max",x_max
y_max=max(data_par%osy,data_par%osy+(data_par%nsy-1)*data_par%dsy)&
+max(data_par%ohy,data_par%ohy+(data_par%nhy-1)*data_par%dhy)
!y_max=data_par%ohy+(data_par%nhy-1)*data_par%dhy

tmp_plane_nx=(x_max-tmp_x_min)/abs(data_par%dhx)+1
write(*,*) "tmp_palne_nx",tmp_plane_nx
tmp_plane_ny=(y_max-tmp_y_min)/abs(data_par%dhy)+1
call from_param('plane_nx',plane_nx,tmp_plane_nx)
call from_param('plane_ny',plane_ny,tmp_plane_ny)

call init_sep3d(plane_data, "OUTPUT","COMPLEX", "REGULAR", ctag="Plane")
call axis_allocate(plane_data,5)
plane_data%o(1)=x_min;       plane_data%d(1)=abs(data_par%dhx);plane_data%n(1)=plane_nx
plane_data%o(2)=y_min;       plane_data%d(2)=abs(data_par%dhy);plane_data%n(2)=plane_ny
plane_data%o(3)=data_par%ow; plane_data%d(3)=data_par%dw; plane_data%n(3)=data_par%nw
plane_data%o(4)=opx;         plane_data%d(4)=dpx;         plane_data%n(4)=npx
plane_data%o(5)=opy;         plane_data%d(5)=dpy;         plane_data%n(5)=npy
call sep3d_set_sep3d(plane_data)
call sep3d_write_description("Plane",plane_data)

call auxputch ("plane_os_x","f",plane_osx,"Plane")
call auxputch ("plane_ds_x","f",plane_dsx,"Plane")
call auxputch ("plane_ns_x","i",plane_nsx,"Plane")
call auxputch ("plane_os_y","f",plane_osy,"Plane")
call auxputch ("plane_ds_y","f",plane_dsy,"Plane")
call auxputch ("plane_ns_y","i",plane_nsy,"Plane")

allocate(wld(plane_nx,plane_ny,data_par%nw))
write(*,*) "f1",plane_nx,plane_ny,data_par%nw
blocksize=plane_nx*plane_ny*data_par%nw*8
do ipy=1,npy
  py=(ipy-1)*dpy+opy
  do ipx=1,npx
    !write(*,*) "f2",ipx
    px=(ipx-1)*dpx+opx
    wld=0.0
    call plane_transfer(x_min,y_min,plane_nx,plane_ny,px,py,wld,&
     plane_osx,plane_dsx,plane_nsx,plane_osy,plane_dsy,plane_nsy)
    !write(*,*) "f3"
    if (srite("Plane",wld,blocksize)/=blocksize ) call seperr("err write")
  end do
end do

deallocate(wld)
end subroutine plane_write



subroutine plane_transfer(min_x,min_y,nx,ny,px,py,wld,&
plane_osx,plane_dsx,plane_nsx,plane_osy,plane_dsy,plane_nsy)
real :: min_x,min_y,px,py
integer :: nx,ny
complex :: wld(:,:,:)
real :: plane_osx,plane_dsx,plane_osy,plane_dsy
integer :: plane_nsx,plane_nsy

complex,allocatable:: tmp_wld(:,:,:)


real :: x_dt,y_dt,t_y,t_x
integer :: blocksize,i_block
integer :: plane_isy,plane_isx,isx,isy,iw,i_ox,i_oy
integer :: i_ox_min, i_ox_max,i_oy_min,i_oy_max,i_x,i_y
real :: plane_sy,shot_oy,plane_sx,shot_ox,w
integer, external:: sreed,sseek_block 

! plane_os_x: first point shot to construct plane source

x_dt=plane_dsx*px
y_dt=plane_dsy*py

write(0,*) "x_dt",x_dt,nx,ny
allocate(tmp_wld(data_par%nhx,data_par%nhy,data_par%nw))

blocksize=8*data_par%nhx*data_par%nhy*data_par%nw
!write(*,*)"plane_ns_y",plane_nsy,plane_nsx
do plane_isy=1,plane_nsy
  t_y=(plane_isy-1)*y_dt
  plane_sy=(plane_isy-1)*plane_dsy+plane_osy
  isy=(plane_sy-data_par%osy)/data_par%dsy+1 ! index in shot gather data
  shot_oy=plane_sy+data_par%ohy
!  shot_oy=data_par%ohy
  i_oy=(shot_oy-min_y)/data_par%dhy+1    ! index in plane wave
  !write(*,*)"i_oy", i_oy
  do plane_isx=1,plane_nsx
    plane_sx=(plane_isx-1)*plane_dsx+plane_osx  
    isx=(plane_sx-data_par%osx)/data_par%dsx+1
    t_x=(plane_isx-1)*x_dt
    shot_ox=plane_sx+data_par%ohx
!    shot_ox=data_par%ohx
    i_ox=(shot_ox-min_x)/abs(data_par%dhx)+1
    !!!Read a shot at :isx,isy  to tmp_wld(x,y,w)
    i_block=(isy-1)*data_par%nsx+isx-1
    if ( sseek_block("Data",i_block,blocksize,0) /=i_block) call seperr("plane source data seek wrong")
    if ( sreed("Data",tmp_wld,blocksize )/= blocksize) call seperr("plane source data read wrong")
    !!! wld(i_ox:i_ox+nhx,i_oy:i_oy+nhy,w)+=tmp_wld*cmplx(cos(w*t),sin(w*t))
    i_ox_min=min(i_ox,i_ox+sign(data_par%dhx)*(data_par%nhx-1))
    i_ox_max=max(i_ox,i_ox+sign(data_par%dhx)*(data_par%nhx-1))
    i_oy_min=min(i_oy,i_oy+sign(data_par%dhy)*(data_par%nhy-1))
    i_oy_max=max(i_oy,i_oy+sign(data_par%dhy)*(data_par%nhy-1))
    do iw=1,data_par%nw
      w=(iw-1)*data_par%dw+data_par%ow
      if (data_par%dhx >0) then
        wld(i_ox_min:i_ox_max,i_oy_min:i_oy_max,iw)=&
          wld(i_ox_min:i_ox_max,i_oy_min:i_oy_max,iw) +&
          tmp_wld(:,:,iw)*cmplx(cos(w*t_x),sin(w*t_x))
      else
        do i_y=1,data_par%nhy
          do i_x=1,data_par%nhx
            wld(i_ox-i_x+1,i_y+i_oy-1,iw)=wld(i_x-i_x+1,i_y+i_oy-1,iw)+&
            tmp_wld(i_x,i_y,iw)*cmplx(cos(w*t_x),sin(w*t_x))
          end do
        end do
      end if
    end do
  end do
end do
deallocate(tmp_wld)
end subroutine plane_transfer

subroutine data_mod_release
deallocate(wavelet)
end subroutine data_mod_release

integer function sign(a)
real :: a
if (a>0) sign=1
if (a==0) sign=0
if (a <0) sign=-1
return
end function sign


end module data_mod
