module shot_kernal_mod
use ani

!parameters for shot_phase_shift
implicit none
type ker_par_type
  integer :: op
  integer :: nx,ny,nref
  real    :: dx,dy,dz
  real    :: fkx,dkx,fky,dky   ! for phase-shift
  real    :: ax,ay,bx,by,tx,ty ! for finite difference
  logical :: is2d
end type ker_par_type

type(ker_par_type),private:: ker_par

integer,private:: m,n
real,private :: ep,dl,phi,f
logical :: llll
!for finite difference
complex,allocatable :: rax(:),rbx(:),rcx(:),ux(:),vx(:)
real,allocatable    :: cax(:),cbx(:)
complex,allocatable :: ray(:),rby(:),rcy(:),uy(:),vy(:)
real,allocatable    :: cay(:),cby(:)
complex,allocatable :: conapp(:),conapm(:),conamp(:),conamm(:)
!end for finite difference


! for mult-ref-vel
real, allocatable :: v0_z(:)
complex,allocatable :: add_wld(:,:),tmp_wld(:,:)

! end for mult-ref-vel


contains

subroutine ker_p_init(op,nx,ny,dx,dy,dz,a,b,trick,is2d)
integer :: op,nx,ny
real    :: dx,dy,dz,a,b,trick
logical ::is2d

real     :: pi
llll=.true.
pi=3.1415926

phi=pi/6.0; f=1.0-15.0*15.0/(27.0*27.0); ep=0.4; dl=0.2
m=nx/2; n=19

allocate(conapp(n),conapm(n),conamp(n),conamm(n))
ker_par%op=op
ker_par%nx=nx; ker_par%ny=ny
ker_par%dx=abs(dx); ker_par%dy=abs(dy); ker_par%dz=dz
ker_par%fkx=-pi/ker_par%dx; ker_par%fky=-pi/ker_par%dy
ker_par%dkx=2.0*pi/(nx*ker_par%dx); ker_par%dky=2.0*pi/(ny*ker_par%dy)
ker_par%ax=a*dz/(2.0*ker_par%dx*ker_par%dx); ker_par%ay=a*dz/(2.0*ker_par%dy*ker_par%dy)
ker_par%bx=b/(ker_par%dx*ker_par%dx); ker_par%by=b/(ker_par%dy*ker_par%dy)
ker_par%tx=trick;     ker_par%ty=trick
ker_par%is2d=is2d
!write(*,*) "-----ker_p_init--------";
!write(*,*) "operator",ker_par%op,"nx",ker_par%nx,"ny",ker_par%ny, &
!           "dx,dy,dz",ker_par%dx,ker_par%dy,ker_par%dz, &
!           "dkx,dky",ker_par%dkx,ker_par%dky, &
!           "ax,ay",ker_par%ax,ker_par%ay,"bx,by",ker_par%bx,ker_par%by,&
!           "tx,ty",ker_par%tx,ker_par%ty
!write(*,*)'inode_ker',ker_par%nx,ker_par%dkx
end subroutine ker_p_init


subroutine ker_p_ani_init(w_vp)
real :: w_vp
integer :: im
call weit(phi,ep,dl,w_vp,f,0.0,ker_par%dkx,m)
call ani_init(phi,ep,dl,w_vp,f,ker_par%dx,ker_par%dkx,ker_par%dz,m,n)
call convlv_coe(m,n,conapp,conapm,conamp,conamm)
end subroutine ker_p_ani_init

subroutine shot_ker_init(nx,ny,nref)
integer :: nx,ny,nref
!for mult-ref-vel
ker_par%nref=nref
allocate( add_wld(nx,ny), tmp_wld(nx,ny),v0_z(nref) )
!end for mult_ref-vel


! for finite difference
allocate ( rax(nx),rbx(nx),rcx(nx),ux(nx),vx(nx),cax(nx),cbx(nx) )
allocate ( ray(ny),rby(ny),rcy(ny),uy(ny),vy(ny),cay(ny),cby(ny) )
!end for finite difference

end subroutine shot_ker_init


subroutine shot_ker_depth_onestep(wld_z,v_z,w,signn,op_p)
complex :: wld_z(:,:)
real :: v_z(:,:),w
integer :: signn
integer :: op_p

integer :: op
integer :: nref
integer :: iref,ix,iy,ivx,iiref
real :: wvx,dref,wvx_all
complex,allocatable ::  tmp_fft(:)

real ::w_v0_z
logical :: aa=.true.

allocate ( tmp_fft(ker_par%nx))
nref=ker_par%nref

! op:1:ssf 2:fd 3:ffd
op=ker_par%op


!v0_z(1)=minval(v_z)
!v0_z(nref)=maxval(v_z)

!dref=(v0_z(nref)-v0_z(1))/(nref-1)
dref=(maxval(v_z)-minval(v_z))/(nref+1)
!if (v0_z(1)==v0_z(nref)) nref=1
if (  abs(maxval(v_z)-minval(v_z))   <50 ) then
   nref=1
   v0_z(1)=0.5*maxval(v_z)+0.5*minval(v_z)
else
   v0_z(1)=minval(v_z)
   v0_z(nref)=maxval(v_z)
   dref=(v0_z(nref)-v0_z(1))/(nref-1)
   v0_z(2)=(minval(v_z)+maxval(v_z))*0.5
!  do iref=1,nref
!    v0_z(iref)=iref*dref
!  end do
end if

!v0_z(2)=0.5*v0_z(1)+0.5*v0_z(3)

!v0_z(1)=0.5*v0_z(1)+0.5*v0_z(2)
!v0_z(3)=0.5*v0_z(3)+0.5*v0_z(2)

if (op_p== 2) then
  v0_z(1)=1500.0
end if



if (op == 2) then
   w_v0_z=0.0
   call shot_ker_ssf(wld_z,w_v0_z,w,v_z,ker_par,signn)
   call shot_ker_fd(wld_z,v_z/w,v_z/w,ker_par,signn)
else
  if (ker_par%is2d) then
    tmp_fft=wld_z(:,1)
    call segFFT1D(tmp_fft,ker_par%nx,+1)
    wld_z(:,1)=tmp_fft(:)
  else
  end if
  add_wld=0.0
  do iref=1,nref
    tmp_wld=wld_z
    w_v0_z=w/v0_z(iref)
    if (op_p== 3) then
      call ani_shot_phase_shift(tmp_wld,w_v0_z,ker_par,signn)
    else
      call shot_phase_shift(tmp_wld,w_v0_z,ker_par,signn)
    end if
    if (ker_par%is2d) then
      tmp_fft=tmp_wld(:,1)
      call segFFT1D(tmp_fft,ker_par%nx,-1)
      tmp_wld(:,1)=tmp_fft(:)
    else
    end if
    if (op_p== 3) then
    else
      call shot_ker_ssf(tmp_wld,w_v0_z,w,v_z,ker_par,signn)
      if (op==3) then
        call shot_ker_fd(tmp_wld,v_z/w-v0_z(iref)/w,v_z/w,ker_par,signn)
      end if 
      call convlv(tmp_wld,ker_par)
    end if
    if ( nref==1) then
      add_wld=tmp_wld
    else
      do iy=1,ker_par%ny 
        do ix=1,ker_par%nx
          ivx=(v_z(ix,iy)-v0_z(1))/dref+1
          wvx=(v_z(ix,iy)-(v0_z(1)+(ivx-1)*dref))/dref
          if (v_z(ix,iy) >= v0_z(nref)) then 
            if (iref==nref) add_wld(ix,iy)=add_wld(ix,iy)+tmp_wld(ix,iy) 
          else
            if (v_z(ix,iy) <= v0_z(1)) then
              if (iref==1)  add_wld(ix,iy)=add_wld(ix,iy)+tmp_wld(ix,iy)
            else
              if (ivx==iref) add_wld(ix,iy)=add_wld(ix,iy)+tmp_wld(ix,iy)*(1-wvx)
              if (ivx+1==iref) add_wld(ix,iy)=add_wld(ix,iy)+tmp_wld(ix,iy)*wvx
            end if
          end if
        end do
      end do
    end if
  end do  
  wld_z=add_wld 
end if

deallocate ( tmp_fft)
end subroutine shot_ker_depth_onestep

subroutine convlv(wld_z,sps_p)
type(ker_par_type):: sps_p
complex ::wld_z(:,:)
integer :: nx,ny
complex,allocatable :: tmpwld_z(:,:)
integer :: iy,ix,ic
nx=sps_p%nx; ny=sps_p%ny
allocate(tmpwld_z(nx,ny))

tmpwld_z=wld_z
wld_z=0.0
do iy=1,ny
  do ix=1,nx
    wld_z(ix,iy)=conapp(1)*tmpwld_z(ix,iy)
    do ic=1,n-1
     if (ix+ic <=nx ) then
       wld_z(ix,iy)=wld_z(ix,iy)+conapp(ic+1)*tmpwld_z(ix+ic,iy)
     end if
     if (ix-ic > 0 ) then
       wld_z(ix,iy)=wld_z(ix,iy)+conapm(ic+1)*tmpwld_z(ix-ic,iy) 
     end if
    end do
  end do
end do 

deallocate(tmpwld_z)
end subroutine

subroutine shot_phase_shift(wld_z,w_v0_z,sps_p,signn)
complex :: wld_z(:,:)
real :: w_v0_z
type(ker_par_type):: sps_p
integer :: signn

integer :: nx,ny
real    :: fkx,fky,dkx,dky,dz

integer :: ix,iy
real    :: kx,ky,kz2,kz,phsft
complex :: dphp,dphm

nx=sps_p%nx; ny=sps_p%ny
fkx=sps_p%fkx; dkx=sps_p%dkx; fky=sps_p%fky; dky=sps_p%dky; dz=sps_p%dz

do iy=1,ny
  if (sps_p%is2d) then
    ky=0.0
  else
    ky=fky+(iy-1)*dky
  end if
  do ix=1,nx
    kx=fkx+(ix-1)*dkx
    kz2=(kx*kx+ky*ky)/(w_v0_z*w_v0_z)
    if (kz2 <1) then
      kz2=1.0-kz2
      phsft=-signn*dz*w_v0_z*(sqrt(kz2))
      wld_z(ix,iy)=wld_z(ix,iy)*cmplx(cos(phsft),sin(phsft))
    else
!      if (kz2>2) then
!        kz2=2.0
!      end if
!      kz2=kz2-1
!      phsft=dz*w_v0_z*(sqrt(kz2))
!      wld_z(ix,iy)=wld_z(ix,iy)*exp(-phsft)
!      wld_z(ix,iy)=0.0
    end if     
  end do
end do


!do iy=1,ny
!  do ix=1,nx
!     kx=fkx+(ix-1)*dkx
!     call  dphase(phi,ep,dl,w_v0_z,f,kx,dz,dphp,dphm)
!     wld_z(ix,iy)=wld_z(ix,iy)*dphp 
!  end do
!end do

llll=.false.
end subroutine shot_phase_shift


subroutine ani_shot_phase_shift(wld_z,w_v0_z,sps_p,signn)
complex :: wld_z(:,:)
real :: w_v0_z
type(ker_par_type):: sps_p
integer :: signn

integer :: nx,ny
real    :: fkx,fky,dkx,dky,dz

integer :: ix,iy
real    :: kx,ky,kz2,kz,phsft,x2,x3,x4,ep,dl,phi,f

ep=0.4; dl=0.2; phi=3.1415926/6.0; f=1.0

nx=sps_p%nx; ny=sps_p%ny
fkx=sps_p%fkx; dkx=sps_p%dkx; fky=sps_p%fky; dky=sps_p%dky; dz=sps_p%dz


!write(*,*) "max3",maxval(abs(wld_z))
do iy=1,ny
  if (sps_p%is2d) then
    ky=0.0
  else
    ky=fky+(iy-1)*dky
  end if
  do ix=1,nx
    kx=fkx+(ix-1)*dkx
    !call kzani(phi,ep,dl,w_v0_z,f,kx,x2,kz,x3,x4)
    phsft=dz*kz
!    write(*,*) kz
    if (kz==0) then
      wld_z(ix,iy)=wld_z(ix,iy)*cmplx(cos(phsft),sin(phsft))
      wld_z(ix,iy)=0.0
    else
      wld_z(ix,iy)=wld_z(ix,iy)*cmplx(cos(phsft),sin(phsft))
    end if
 
  end do
end do
!write(*,*) "max4",maxval(abs(wld_z))
end subroutine ani_shot_phase_shift




subroutine shot_ker_ssf(wld_z,w_v0_z,w,v_z,s_ssf_p,signn)
complex :: wld_z(:,:)
real    :: w,v_z(:,:),w_v0_z
type(ker_par_type) :: s_ssf_p
integer :: signn

integer :: nx,ny
real    :: dz

integer :: ix,iy
real    :: phsft

nx=s_ssf_p%nx; ny=s_ssf_p%ny; dz=s_ssf_p%dz

do ix=1,nx
  do iy=1,ny
    phsft=-signn*dz*(w/v_z(ix,iy)-w_v0_z)
    wld_z(ix,iy)=wld_z(ix,iy)*cmplx(cos(phsft),sin(phsft))
  end do
end do
end subroutine shot_ker_ssf


subroutine shot_ker_fd(wld_z,ca,cb,s_fd_p,signn)
complex :: wld_z(:,:)
real    :: ca(:,:),cb(:,:)
type(ker_par_type) :: s_fd_p
integer :: signn

integer :: nx,ny
real    :: tx,ty,ax,ay,bx,by


integer :: ix,iy

ax=-signn*s_fd_p%ax; ay=-signn*s_fd_p%ay; bx=s_fd_p%bx; nx=s_fd_p%nx
by=s_fd_p%by; tx=s_fd_p%tx; ty=s_fd_p%ty; ny=s_fd_p%ny
!do ix=1,nx
!  uy=wld_z(ix,:)
!  cay=ca(ix,:)
!  cby=cb(ix,:)
!  call fd(uy,vy,ty,ay,by,cay,cby,ny,ray,rby,rcy)
!  wld_z(ix,:)=vy
!end do
!write(*,*) "maxval",maxval(real(wld_z))
do iy=1,ny
  ux=wld_z(:,iy)
  cax=ca(:,iy)
  cbx=cb(:,iy)
  call fd(ux,vx,tx,ax,bx,cax,cbx,nx,rax,rbx,rcx)
  wld_z(:,iy)=vx
end do
end subroutine shot_ker_fd

subroutine fd(u,v,tr,a,b,ca,cb,nx,ra,rb,rc)
complex :: u(:),v(:)
complex :: ra(:),rb(:),rc(:)
real    :: ca(:),cb(:)
real    :: a,b,tr
integer :: nx

integer :: ix
complex :: url,urr,ror,rol

do ix=1,nx
  if (ix==1)  then
    url=0.0
  else
    url=u(ix-1)
  end if
  if (ix==nx) then
    urr=0.0
  else
    urr=u(ix+1)
  end if
  ror=cmplx(tr+b*cb(ix)*cb(ix),a*ca(ix))
  rol=conjg(ror)
  v(ix)=u(ix)+ror*(url+urr-2.0*u(ix))
  ra(ix)=1.0-2.0*rol
  if (ix <nx) rb(ix) =rol
  if (ix >1) rc(ix-1)=rol
end do
call thryj(ra,rb,rc,v,nx,1)
end subroutine fd


subroutine thryj(a,b,c,v,h,yJ)
implicit none
integer h,yJ
complex :: a(:),b(:),c(:),v(:)
integer i
do i=1,yJ
   b(i)=b(i)/a(i)
   v(i)=v(i)/a(i)
enddo
do i=yJ+1,h
   a(i)=a(i)-b(i-yJ)*c(i-yJ)
   v(i)=(v(i)-c(i-yJ)*v(i-yJ))/a(i)
   if(i.le.h-yJ) then
      b(i)=b(i)/a(i)
   endif
enddo
do i=h-yJ,1,-1
  v(i)=v(i)-b(i)*v(i+yJ)
enddo
end subroutine thryj


subroutine shot_ker_release
deallocate(tmp_wld,add_wld,v0_z)
deallocate(rax,rbx,rcx,cax,cbx,ux,vx)
deallocate(ray,rby,rcy,cay,cby,uy,vy)
deallocate(conapp,conapm,conamp,conamm)
call ani_release
end subroutine shot_ker_release

end module shot_kernal_mod
