module math_mod
contains
subroutine qrdcmp(a,m,n,d,sing)
! a(m,n)
integer :: n,m,sing
real :: scale,sigma,sum,tau
real :: a(:,:),d(:)
real,allocatable :: c(:)

allocate(c(n))
scale=0.0
sing=0
do k=1,n
  do i=k,m
    if (abs(a(i,k))>scale) scale=abs(a(i,k))
  end do
  if (scale==0.0) then
    sing=1
    c(k)=0.0; d(k)=0.0
  else
    do i=k,m
       a(i,k)=a(i,k)/scale
    end do
    sum=0.0

    do i=k,m
      sum=sum+a(i,k)*a(i,k)
    end do

    if (a(k,k)>0) then
       sigma=sqrt(sum)
    else
       sigma=-sqrt(sum)
    end if
    a(k,k)=a(k,k)+sigma
    c(k)=sigma*a(k,k)
    d(k)=-scale*sigma

    do j=k+1,n
      sum=0.0
      do i=k,m
        sum=sum+a(i,k)*a(i,j)
      end do
      tau=sum/c(k)
      do i=k,m
        a(i,j)=a(i,j)-tau*a(i,k)
      end do
    end do
  end if
end do
deallocate(c)
end subroutine qrdcmp

subroutine qrdpost(a,m,n,d,q,r)
integer :: m,n
real :: a(:,:),q(:,:),r(:,:),d(:)
real,allocatable :: qt(:,:)  


integer :: k,l,i,j
real :: con

allocate(qt(m,m))

do k=1,n
  do l=1,n
    if (l>k) then
      r(k,l)=a(k,l)
    else
      if (l<k) then
        r(k,l)=0.0
      else
        r(k,l)=d(k)
      end if
    end if
  end do
end do


do k=1,m
   do l=1,m
      if (l==k) then
         q(k,l)=1.0
      else
         q(k,l)=0.0
      end if
   end do
end do

do i=n,1,-1
  con=0.0
  do k=i,m
    con=con+a(k,i)*a(k,i)
  end do
  con=con/2.0

  do k=i,m
    do l=i,n
      qt(k,l)=0.0
      do j=i,m
        qt(k,l)=qt(k,l)+q(j,l)*a(k,i)*a(j,i)/con
      end do
    end do
  end do


  do k=i,m
    do l=i,m
      q(k,l)=q(k,l)-qt(k,l)
    end do
  end do

end do


!do i=n,1,-1
!  con=0.0
!  do k=i,m
!    con=con+a(k,i)*a(k,i)
!  end do
!  con=con/2.0

!  do k=i,m
!    do l=i,m
!      qt(k,l)=0.0
!      do j=i,m
!        qt(k,l)=qt(k,l)+q(j,l)*a(k,i)*a(j,i)/con
!      end do
!    end do
!  end do


!  do k=i,m
!    do l=i,m
!      q(k,l)=q(k,l)-qt(k,l)
!    end do
!  end do

!end do



deallocate(qt)

end subroutine






subroutine qrcon(q,r,dpp,conap,m,n)
real :: q(:,:),r(:,:)
complex :: dpp(:),conap(:)
integer :: i_n,i_m
do i_n=1,n
  conap(i_n)=0.0
  do i_m=1,m
    conap(i_n)=conap(i_n) +q(i_m,i_n)*dpp(i_m)
  end do
end do
!call downtrisolver(r,conap,n)
call uptrisolver(r,conap,n) 

end subroutine qrcon

subroutine downtrisolver(r,conap,n)
real:: r(:,:)
complex:: conap(:)
integer ::n
integer :: i_n,i_m

do i_n=1,n
  do i_m=1,i_n-1
    conap(i_n)=conap(i_n)-conap(i_m)*r(i_m,i_n)
  end do
  conap(i_n)=conap(i_n)/r(i_n,i_n)
end do

end subroutine downtrisolver


subroutine uptrisolver(r,conap,n)
real:: r(:,:)
complex:: conap(:)
integer ::n

integer :: i_n,i_m

do i_n=n,1,-1
  do i_m=i_n+1,n
    conap(i_n)=conap(i_n)-conap(i_m)*r(i_n,i_m)
  end do
  conap(i_n)=conap(i_n)/r(i_n,i_n)
end do

end subroutine uptrisolver



subroutine e4solver(a0,a1,a2,a3,a4,x1,x2,x3,x4,f,w_vp,kx)
real :: a0,a1,a2,a3,a4,f,w_vp,kx
real :: x1,x2,x3,x4

complex :: y1,y2,y3
real :: yr1,yr2,yr3,y01,y02,y03
real :: b0,b1,b2,b3
real :: f41,f42,f43,test

a3=a3/a4; a2=a2/a4; a1=a1/a4; a0=a0/a4; a4=1
!write(*,*)a0,a1,a2,a3,a4
  call coe_3from4(a0,a1,a2,a3,a4,b0,b1,b2,b3)
  dett=detm(a0,a1,a2,a3)
!  write(*,*)"dettttttttttttttttt",dett
  if (dett >0.0) then
     call e3solver(b0,b1,b2,b3,y1,y2,y3)
!     write(*,*) y1,y2,y3
     yr1=y1; yr2=y2; yr3=y3
     call inorder(yr1,yr2,yr3,y01,y02,y03)
     f41=poly4(a0,a1,a2,a3,a4,y01)
     f42=poly4(a0,a1,a2,a3,a4,y02)   
     f43=poly4(a0,a1,a2,a3,a4,y03)
     test=w_vp**2-kx**2
     if (test >0 ) then
       x1=sqrt(w_vp**2-kx**2); x2=-sqrt(w_vp**2-kx**2);
       call newton_p4(a0,a1,a2,a3,a4,x1)
       call newton_p4(a0,a1,a2,a3,a4,x2)
     else
       x1=0.0; x2=0.0
     end if
!     if (abs(x1)>sqrt(w_vp**2-kx**2)) then
!         x1=0.0
!         write(0,*)"nnnnnnnnn"
!     end if
!     if (abs(x2)>sqrt(w_vp**2-kx**2)) then
!         x2=0.0
!         write(0,*)"nnnnnnnnn"
!     end if
!     if (f41<0 .and. f42 >0 )then
!        x1=0.5*(y01+y02)
!        call newton_p4(a0,a1,a2,a3,a4,x1)
!     else
!        x1=0.0
!     end if
!     if (f43 <0 .and. f42>0 ) then
!       x2=0.5*(y02+y03)
!       call newton_p4(a0,a1,a2,a3,a4,x2)
!     else
!        x2=0.0
!     end if
  else 
    !write(0,*)"abcdefg",kx 
    x1=sqrt(w_vp**2-kx**2); x2=-sqrt(w_vp**2-kx**2);
    call coe_3from4(a0,a1,a2,a3,a4,b0,b1,b2,b3)
    yr1=0.0
    call e3solver_numerical(b0,b1,b2,b3,yr1)
    if ( poly4(a0,a1,a2,a3,a4,yr1) > 0 )then
      !write(0,*) "errorerror"
      x1=0.0; x2=0.0
    else 
      call newton_p4(a0,a1,a2,a3,a4,x1)
      call newton_p4(a0,a1,a2,a3,a4,x2)
    end if
    if (abs(w_vp)<abs(kx)) then 
       x1=0.0; x2=0.0
    else
       if (abs(x1)>sqrt(w_vp**2-kx**2)) then
         x1=0.0
       end if
       if (abs(x2)>sqrt(w_vp**2-kx**2)) then
         x2=0.0
       end if
    end if
  end if
  
end subroutine e4solver

subroutine chooseproot(z1,z2,z3,z4,x1,x2)
complex :: z1,z2,z3,z4,x1,x2
complex :: z(4),xx1,xx2,tmpxx,zz(2)
integer ::j
real :: absi,absx
z(1)=z1; z(2)=z2; z(3)=z3; z(4)=z4
j=0
do i=1,4
   if (abs(aimag(z(i))) < 0.00001*abs(real(z(i)))) then
     j=j+1
   end if
end do

if (j==0 .or. j==2) then
! no real root
  absi=abs(aimag(z(1)))
  do i=1,4
    if (absi <abs(aimag(z(i))) ) then
       absi=abs(aimag(z(i)))
    end if
  end do
  k=0
  do i=1,4
    if (absi==abs(aimag(z(i))) ) then
       k=k+1
       zz(k)=z(i)
    end if
  end do
end if

if (j==4) then
! 4 real root
  absx=abs( real(z(1)))
  xx1=z(1)
  do i=1,4
    if (absx >abs( real(z(i))) ) then
      absx=abs(real(z(i)))
      xx1=z(i)
    end if
  end do
  if ( xx1==z(1) ) then
    absx=abs(real(z(2)))
    xx2=z(2)
  else
    absx=abs(real(z(1)))
    xx2=z(1)
  end if
  do i=1,4
    if (absx > abs(real(z(i))) .and. xx1/=z(i) ) then
      absx=abs(real(z(i)))
      xx2=z(i)
    end if
  end do 
  if (real(xx1) < 0 ) then
    tmpxx=xx1
    xx1=xx2
    xx2=tmpxx
  end if
  zz(1)=xx1
  zz(2)=xx2
end if
x1=zz(1); x2=zz(2)

end subroutine

subroutine p4solver(a0,a1,a2,a3,a4,z1,z2,z3,z4)
real :: a0,a1,a2,a3,a4
real :: x1,x2,x3,x4
complex :: y(3)
real :: b3,b2,b1,b0
real :: ctmp1,ctmp2,yy
complex :: z1,z2,z3,z4,Dtmp,Rtmp,Etmp

a0=a0/a4; a1=a1/a4; a2=a2/a4; a3=a3/a4; a4=1.0
b3=1; b2=-a2; b1=a1*a3-4.0*a0; b0=4.0*a2*a0-a1**2-a3**2*a0
call e3solver(b0,b1,b2,b3,y(1),y(2),y(3))
do i=1,3
  if (abs(aimag(y(i)))< abs(real(y(i)))*0.00001) then
    !write(*,*) i
    yy=real(y(i))
    ctmp1=a3**2-4.0*a2+4.0*yy
    Rtmp=0.5*csqrt(ctmp1)
    ctmp2=yy**2-4.0*a0
    if (Rtmp==0.0) then
       Dtmp=sqrt( 3.0/4.0*a3**2 - 2.0*a2+2.0*sqrt(ctmp2) )
    else
       Dtmp=csqrt( 3.0/4.0*a3**2-Rtmp**2-2.0*a2+0.25*(4.0*a3*a2-8.0*a1-a3**3)/Rtmp)
    end if

    if (Rtmp==0.0) then
       Etmp=sqrt( 3.0/4.0*a3**2 - 2.0*a2- 2.0*sqrt(ctmp2) )
    else
       Etmp=csqrt( 3.0/4.0*a3**2-Rtmp**2-2.0*a2-0.25*(4.0*a3*a2-8.0*a1-a3**3)/Rtmp)
    end if
    z1=-0.25*a3+0.5*Rtmp+0.5*Dtmp
    z2=-0.25*a3+0.5*Rtmp-0.5*Dtmp 
    z3=-0.25*a3-0.5*Rtmp+0.5*Etmp
    z4=-0.25*a3-0.5*Rtmp-0.5*Etmp
    exit
  end if
end do


if (abs(aimag(z1)) < 0.00001*abs(real(z1)) ) then
  x1=real(z1)
else
  x1=0.0
end if
if (abs(aimag(z2)) < 0.00001*abs(real(z2)) ) then
  x2=real(z2)
else
  x2=0.0
end if
if (abs(aimag(z3)) < 0.00001*abs(real(z3)) ) then
  x3=real(z3)
else
  x3=0.0
end if
if (abs(aimag(z4)) < 0.00001*abs(real(z4)) ) then
  x4=real(z4)
else
  x4=0.0
end if
!write(*,*) z1,z2,z3,z4
end subroutine



subroutine e3solver(a0,a1,a2,a3,x1,x2,x3)
real::a0,a1,a2,a3
complex:: x1,x2,x3
real:: p,q
real ::pi
complex::w3,w,w1,w2

pi=3.1415926
a2=a2/a3; a1=a1/a3; a0=a0/a3
p=(3.0*a1-a2*a2)/3.0
q=(9.0*a1*a2-27*a0-2.0*a2**3  )/27.0
aa=0.25*q*q+1.0/27.0*p**3
w3=0.5*q+csqrt(aa)
w=(w3)**(1.0/3.0)
x1=w-p/(3.0*w)-1.0/3.0*a2
w1=w*cmplx(cos(pi*2.0/3.0),sin(pi*2.0/3.0))
x2=w1-p/(3.0*w1)-1.0/3.0*a2
w2=w*cmplx(cos(pi*4.0/3.0),sin(pi*4.0/3.0))
x3=w2-p/(3.0*w2)-1.0/3.0*a2
end subroutine e3solver

real function detm(a0,a1,a2,a3)
detm=a1**2*a2**2-4.0*a0*a2**3-4.0*a1**3*a3+18.0*a0*a1*a2*a3-27*a0**2*a3**2
end function detm

subroutine coe_3from4(a0,a1,a2,a3,a4,b0,b1,b2,b3)
real :: a0,a1,a2,a3,a4
real :: b0,b1,b2,b3

b3=4.0*a4; b2=3.0*a3; b1=2.0*a2; b0=a1
end subroutine

subroutine e3solver_numerical(b0,b1,b2,b3,y)
real :: b0,b1,b2,b3,y
real :: c0,c1,c2
integer ::i
real :: p,f3

c2=3.0*b3; c1=2.0*b2; c0=b1
do i=1,20
  p=poly2(c0,c1,c2,y)   
  f3=poly3(b0,b1,b2,b3,y)
  y=y-f3/p
end do 
end subroutine e3solver_numerical

subroutine newton_p4(a0,a1,a2,a3,a4,x)
real :: a0,a1,a2,a3,a4,x
real :: b0,b1,b2,b3,p,f4
integer :: i
call coe_3from4(a0,a1,a2,a3,a4,b0,b1,b2,b3)
do i=1,30
 p=poly3(b0,b1,b2,b3,x)
 f4=poly4(a0,a1,a2,a3,a4,x)
 x=x-f4/p
end do
end subroutine newton_p4

subroutine inorder(yr1,yr2,yr3,y01,y02,y03)
real :: yr1,yr2,yr3
real :: y01,y02,y03
y01=0
y01=yr1
if (yr2>y01) y01=yr2
if (yr3>y01) y01=yr3

y03=yr3
if (yr1 <y03) y03=yr1
if (yr2 <y03) y03=yr2

if (yr1> y03 .and. yr1 <y01) y02=yr1
if (yr2> y03 .and. yr2 <y01) y02=yr2
if (yr3> y03 .and. yr3 <y01) y02=yr3
end subroutine inorder


real function poly4(a0,a1,a2,a3,a4,x)
real :: a0,a1,a2,a3,a4,x
poly4=a0+a1*x+ a2*x*x+ a3*x*x*x+ a4*x*x*x*x
end function poly4

real function poly3(a0,a1,a2,a3,x)
real :: a0,a1,a2,a3
poly3=a0+a1*x+ a2*x*x+ a3*x*x*x
end function poly3


real function poly2(a0,a1,a2,x)
real :: a0,a1,a2,x
poly2=a0+a1*x+ a2*x*x
end function poly2



end module math_mod
