module sinc_f90_mod
contains

subroutine sinc_f90_mksinc(d,lsinc,sinc)
real :: d
integer :: lsinc
real :: sinc(:)

integer :: j
real :: s(20),a(20),c(20),work(20),fmax
!compute auto-correlation and cross-correlation arrays
 fmax = 0.066+0.265 *aLog(lsinc)
 if (fmax >=1.0) then
   fmax=1.0
 end if
 do j=0,lsinc-1
    a(j+1)=sinc_f90_sinc(fmax*(j))
    c(j+1)=sinc_f90_sinc(fmax*(lsinc/2-j-1+d))
 end do
 
 !solve symmetric Toeplitz system for the sinc approximation
 call sinc_f90_stoepd(lsinc,a,c,s,work)
 do j=1,lsinc
   sinc(j)=s(j)
 end do
end subroutine sinc_f90_mksinc

real function sinc_f90_sinc(x)
real :: x
real :: pix
pix=x*3.1415926

if (x == 0.0) then
   sinc_f90_sinc=1.0
else
   sinc_f90_sinc=sin(pix)/pix
end if
return
end function sinc_f90_sinc


subroutine sinc_f90_stoepd(n,r,g,f,a)
integer ::n
real :: r(:),g(:),f(:),a(:)

integer :: i,j
real :: v,e,c,w,bot

if (r(1) == 0.0) return
a(1)=1.0
v=r(1)
f(1)=g(1)/r(1)

do j=1,n-1
  !solve Ra=v as in Claerbout, FGDP, p. 57
  a(j+1)=0.0
  f(j+1)=0.0
  e=0.0
  do i=0,j-1
    e=e+a(i+1)*r(j-i+1)
  end do
  c=e/v
  v=v-c*e
  do i=0,j/2
    bot=a(j-i+1)-c*a(i+1)
    a(i+1)=a(i+1)-c*a(j-i+1)
    a(j-i+1)=bot
  end do
  ! use a and v above to get f[i], i = 0,1,2,...,j
  w=0.0
  do i=0,j-1
    w=w+f(i+1)*r(j-i+1)
  end do
  c=(w-g(j+1))/v
  do i=0,j
    f(i+1)=f(i+1)-c*a(j-i+1)
  end do
end do

end subroutine sinc_f90_stoepd

end module sinc_f90_mod
