module velocity_mod
use sep3d_struct_mod
use mpi_sep

implicit none

type vel_par_type
  real :: dx,dy,dz
  real :: min_x,min_y,min_z
  real :: mig_dx,mig_dy,mig_dz
  integer :: nx,ny,nz
  integer :: MPI_size
end type vel_par_type

type(sep3d),private :: velocity,ep,dl
type(vel_par_type),private :: vel_par,ep_par,dl_par

contains
subroutine velocity_mod_init(mig_dx,mig_dy,mig_dz,is2d,i_node)
real :: mig_dx,mig_dy,mig_dz
integer ::i_node
logical :: is2d

integer,external ::auxin,sreed
integer :: ierr,size_real,size_int

call MPI_TYPE_SIZE(MPI_REAL,size_real,ierr)
call MPI_TYPE_SIZE(MPI_INTEGER,size_int,ierr)
vel_par%MPI_size=9*size_real+4*size_int

if (i_node==0) then
  if (-1==auxin("Velocity")) then
    call seperr("Must provide Velocity file")
  end if
  call init_sep3d("Velocity", velocity, "INPUT")
  vel_par%dz=velocity%d(1); vel_par%nz=velocity%n(1);vel_par%min_z=velocity%o(1)
  vel_par%dx=velocity%d(2); vel_par%nx=velocity%n(2);vel_par%min_x=velocity%o(2)
  if (size(velocity%n)<2 .or. size(velocity%n)>3) then
     call seperr("Velocity file should 2D or 3D")
  else
    if (size(velocity%n)==2) then
       vel_par%dy=10; vel_par%ny=1; vel_par%min_y=0 
       is2d=.true.
    else
       vel_par%dy=velocity%d(3); vel_par%ny=velocity%n(3)
       vel_par%min_y=velocity%o(3)
       is2d=.false.
    end if 
  end if
  vel_par%mig_dx=abs(mig_dx); vel_par%mig_dy=abs(mig_dy); vel_par%mig_dz=mig_dz

!  if (-1==auxin("EP")) then
!    call seperr("Must provide EP file")
!  end if
!  call init_sep3d("EP", ep, "INPUT")
!  ep_par%dz=ep%d(1); ep_par%nz=ep%n(1);ep_par%min_z=ep%o(1)
!  ep_par%dx=ep%d(2); ep_par%nx=ep%n(2);ep_par%min_x=ep%o(2)
!  ep_par%dy=10;      ep_par%ny=1;      ep_par%min_y=0 
!  ep_par%mig_dx=abs(mig_dx); ep_par%mig_dy=abs(mig_dy); ep_par%mig_dz=mig_dz  

!  if (-1==auxin("DL")) then
!    call seperr("Must provide DL file")
!  end if
!  call init_sep3d("DL", dl, "INPUT")
!  dl_par%dz=dl%d(1); dl_par%nz=dl%n(1);dl_par%min_z=dl%o(1)
!  dl_par%dx=dl%d(2); dl_par%nx=dl%n(2);dl_par%min_x=dl%o(2)
!  dl_par%dy=10;      dl_par%ny=1;      dl_par%min_y=0 
!  dl_par%mig_dx=abs(mig_dx); dl_par%mig_dy=abs(mig_dy); dl_par%mig_dz=mig_dz

end if
call MPI_BCAST(vel_par,vel_par%MPI_Size,MPI_BYTE,0,MPI_COMM_WORLD,ierr)
!write(*,*) vel_par%nx," velocity ",vel_par%ny
end subroutine velocity_mod_init

subroutine ep_read(min_x,min_y,min_z,nx,ny,nz,epi,is2d)
real :: epi(:,:,:)
real :: min_x,min_y,min_z
integer :: nx,ny,nz
logical :: is2d

real :: max_x,max_y
real,allocatable :: tmp_vel(:,:,:)
integer :: ix,iy,iz
integer :: min_ix,max_ix,min_iy,max_iy
integer :: tmp_vnx, tmp_vny,tmp_iy,tmp_ix,tmp_iz
integer  :: n_block
real :: tmp_min_x,tmp_min_y
real    :: tmp_wx,tmp_wy, tmp_wz,tmp_x,tmp_y,tmp_z,dtmp_x,dtmp_z
real    :: w11,w12,w21,w22
real    :: v11,v12,v21,v22
integer :: tmp_iy_p_1
integer,external:: sreed,sseek_block

  max_x=min_x+(nx-1)*ep_par%mig_dx; max_y=min_y+(ny-1)*ep_par%mig_dy

  min_ix=(min_x-ep_par%min_x)/ep_par%dx+1;  max_ix=(max_x-ep_par%min_x)/ep_par%dx+1
  min_iy=(min_y-ep_par%min_y)/ep_par%dy+1;  max_iy=(max_y-ep_par%min_y)/ep_par%dy+1

  if (min_ix <1 ) min_ix=1
  if (max_ix > ep_par%nx ) max_ix=ep_par%nx
  if (min_iy <1 ) min_iy=1
  if (max_iy > ep_par%ny ) max_iy=ep_par%ny

  !write(*,*)"minx",min_x,"maxx",max_x,"miny",min_y,"maxy",max_y
  !write(*,*)"minix",min_ix,"maxix",max_ix,"miniy",min_iy,"maxiy",max_iy
  tmp_min_x=(min_ix-1)*ep_par%dx+ ep_par%min_x; tmp_vnx=max_ix-min_ix+1
  tmp_min_y=(min_iy-1)*ep_par%dy+ ep_par%min_y; tmp_vny=max_iy-min_iy+1
  !write(*,*)"velnx",tmp_vnx,"velny",tmp_vny,"nz",vel_par%nz

  allocate(tmp_vel(ep_par%nz,tmp_vnx,tmp_vny) )

  do iy=1,max_iy
     n_block=(iy-1)*ep_par%nx+min_ix-1
     if(sseek_block("EP",n_block,ep_par%nz*4,0)/= n_block) &
      call seperr("velocity seek wrong7")
     if(sreed("EP",tmp_vel(:,:,iy),ep_par%nz*tmp_vnx*4)/=ep_par%nz*tmp_vnx*4) &
      call seperr("velocity read wrong8")
  end do
  do iy=1,ny
    tmp_y=(iy-1)*ep_par%mig_dy+min_y; tmp_y=(tmp_y-tmp_min_y)/ep_par%dy+1.0 
    tmp_iy=int(tmp_y);       tmp_wy=tmp_y-tmp_iy
  
    if (tmp_iy <1) then
      tmp_iy=1;              tmp_wy=0.5
    end if
    if (tmp_iy > tmp_vny ) then
      tmp_iy=tmp_vny;tmp_wy=0.5
    end if
    if (tmp_iy< tmp_vny) then
      tmp_iy_p_1=tmp_iy+1
    else
      tmp_iy_p_1=tmp_iy
    end if

    do ix=1,nx
      tmp_x=(ix-1)*ep_par%mig_dx+min_x; tmp_x=(tmp_x-tmp_min_x)/ep_par%dx+1.0
      tmp_ix=int(tmp_x);  tmp_wx=tmp_x-tmp_ix
      if (tmp_ix <1) then
        tmp_ix=1;  tmp_wx=0.5
      end if
      if (tmp_ix >= tmp_vnx ) then 
        tmp_ix=tmp_vnx-1;  tmp_wx=0.5
      end if
      w11=(1-tmp_wx)*(1-tmp_wy);w12=(1-tmp_wx)*tmp_wy 
      w21=tmp_wx*(1-tmp_wy);w22=tmp_wx*tmp_wy
      do iz=1,nz
        tmp_z=(iz-1)*ep_par%mig_dz+min_z; tmp_z=(tmp_z-ep_par%min_z)/ep_par%dz+1.0
        tmp_iz=int(tmp_z); tmp_wz=tmp_z-tmp_iz
        if (tmp_iz < 1) then
          tmp_iz=1;   tmp_wz=0.5
        end if
        if (tmp_iz >= ep_par%nz ) then
          tmp_iz=ep_par%nz-1; tmp_wz=0.5
        end if
        v11=tmp_vel(tmp_iz,tmp_ix,tmp_iy)*(1-tmp_wz)&
           +tmp_wz*tmp_vel(tmp_iz+1,tmp_ix,tmp_iy)
        v12=tmp_vel(tmp_iz,tmp_ix,tmp_iy_p_1)*(1-tmp_wz)&
           +tmp_wz*tmp_vel(tmp_iz+1,tmp_ix,tmp_iy_p_1)
        v21=tmp_vel(tmp_iz,tmp_ix+1,tmp_iy)*(1-tmp_wz)&
           +tmp_wz*tmp_vel(tmp_iz+1,tmp_ix+1,tmp_iy)
        v22=tmp_vel(tmp_iz,tmp_ix+1,tmp_iy_p_1)*(1-tmp_wz)&
           +tmp_wz*tmp_vel(tmp_iz+1,tmp_ix+1,tmp_iy_p_1)
        epi(iz,ix,iy)=v11*w11+v12*w12+v21*w21+v22*w22
      end do
    end do
  end do
  write(*,*)"vel read finish"
  deallocate (tmp_vel)

end subroutine ep_read

subroutine velocity_read(min_x,min_y,min_z,nx,ny,nz,vel,is2d,theta,ix_center)
real :: vel(:,:,:)
real :: min_x,min_y,min_z
integer :: nx,ny,nz
logical :: is2d
real :: theta
integer :: ix_center

real :: max_x,max_y
real,allocatable :: tmp_vel(:,:,:)

integer :: ix,iy,iz
integer :: min_ix,max_ix,min_iy,max_iy
integer :: tmp_vnx, tmp_vny,tmp_iy,tmp_ix,tmp_iz
integer  :: n_block
real :: tmp_min_x,tmp_min_y
real    :: tmp_wx,tmp_wy, tmp_wz,tmp_x,tmp_y,tmp_z,dtmp_x,dtmp_z
real    :: w11,w12,w21,w22
real    :: v11,v12,v21,v22
integer :: tmp_iy_p_1
integer,external:: sreed,sseek_block

if (theta==-0.1) then
  max_x=min_x+(nx-1)*vel_par%mig_dx; max_y=min_y+(ny-1)*vel_par%mig_dy

  min_ix=(min_x-vel_par%min_x)/vel_par%dx+1;  max_ix=(max_x-vel_par%min_x)/vel_par%dx+1
  min_iy=(min_y-vel_par%min_y)/vel_par%dy+1;  max_iy=(max_y-vel_par%min_y)/vel_par%dy+1

  if (min_ix <1 ) min_ix=1
  if (max_ix > vel_par%nx ) max_ix=vel_par%nx
  if (min_iy <1 ) min_iy=1
  if (max_iy > vel_par%ny ) max_iy=vel_par%ny

  !write(*,*)"minx",min_x,"maxx",max_x,"miny",min_y,"maxy",max_y
  !write(*,*)"minix",min_ix,"maxix",max_ix,"miniy",min_iy,"maxiy",max_iy
  tmp_min_x=(min_ix-1)*vel_par%dx+ vel_par%min_x; tmp_vnx=max_ix-min_ix+1
  tmp_min_y=(min_iy-1)*vel_par%dy+ vel_par%min_y; tmp_vny=max_iy-min_iy+1
  !write(*,*)"velnx",tmp_vnx,"velny",tmp_vny,"nz",vel_par%nz

  allocate(tmp_vel(vel_par%nz,tmp_vnx,tmp_vny) )

  do iy=1,max_iy
     n_block=(iy-1)*vel_par%nx+min_ix-1
     if(sseek_block("Velocity",n_block,vel_par%nz*4,0)/= n_block) &
      call seperr("velocity seek wrong1")
     if(sreed("Velocity",tmp_vel(:,:,iy),vel_par%nz*tmp_vnx*4)/=vel_par%nz*tmp_vnx*4) &
      call seperr("velocity read wrong2")
  end do

  do iy=1,ny
    tmp_y=(iy-1)*vel_par%mig_dy+min_y; tmp_y=(tmp_y-tmp_min_y)/vel_par%dy+1.0 
    tmp_iy=int(tmp_y);       tmp_wy=tmp_y-tmp_iy
  
    if (tmp_iy <1) then
      tmp_iy=1;              tmp_wy=0.5
    end if
    if (tmp_iy > tmp_vny ) then
      tmp_iy=tmp_vny;tmp_wy=0.5
    end if
    if (tmp_iy< tmp_vny) then
      tmp_iy_p_1=tmp_iy+1
    else
      tmp_iy_p_1=tmp_iy
    end if
 
 
    do ix=1,nx
      tmp_x=(ix-1)*vel_par%mig_dx+min_x; tmp_x=(tmp_x-tmp_min_x)/vel_par%dx+1.0
      tmp_ix=int(tmp_x);  tmp_wx=tmp_x-tmp_ix
      if (tmp_ix <1) then
        tmp_ix=1;  tmp_wx=0.5
      end if
      if (tmp_ix >= tmp_vnx ) then 
        tmp_ix=tmp_vnx-1;  tmp_wx=0.5
      end if
      w11=(1-tmp_wx)*(1-tmp_wy);w12=(1-tmp_wx)*tmp_wy 
      w21=tmp_wx*(1-tmp_wy);w22=tmp_wx*tmp_wy
      do iz=1,nz
        tmp_z=(iz-1)*vel_par%mig_dz+min_z; tmp_z=(tmp_z-vel_par%min_z)/vel_par%dz+1.0
        tmp_iz=int(tmp_z); tmp_wz=tmp_z-tmp_iz
        if (tmp_iz < 1) then
          tmp_iz=1;   tmp_wz=0.5
        end if
        if (tmp_iz >= vel_par%nz ) then
          tmp_iz=vel_par%nz-1; tmp_wz=0.5
        end if
        v11=tmp_vel(tmp_iz,tmp_ix,tmp_iy)*(1-tmp_wz)&
           +tmp_wz*tmp_vel(tmp_iz+1,tmp_ix,tmp_iy)
        v12=tmp_vel(tmp_iz,tmp_ix,tmp_iy_p_1)*(1-tmp_wz)&
           +tmp_wz*tmp_vel(tmp_iz+1,tmp_ix,tmp_iy_p_1)
        v21=tmp_vel(tmp_iz,tmp_ix+1,tmp_iy)*(1-tmp_wz)&
           +tmp_wz*tmp_vel(tmp_iz+1,tmp_ix+1,tmp_iy)
        v22=tmp_vel(tmp_iz,tmp_ix+1,tmp_iy_p_1)*(1-tmp_wz)&
           +tmp_wz*tmp_vel(tmp_iz+1,tmp_ix+1,tmp_iy_p_1)
        vel(iz,ix,iy)=v11*w11+v12*w12+v21*w21+v22*w22
      end do
    end do
  end do
  write(*,*)"vel read finish"
  deallocate (tmp_vel)
else
  if (theta >= 0) then
    allocate(tmp_vel(vel_par%nz,vel_par%nx,1))
    tmp_min_x=vel_par%min_x
    tmp_vnx=vel_par%nx
  
    do iy=1,ny
      n_block=(iy-1)*vel_par%nx
      if (sseek_block("Velocity",n_block,vel_par%nz*4,0)/= n_block) call seperr("velocity seek wrong3")
      if(sreed("Velocity",tmp_vel(:,:,1),vel_par%nz*vel_par%nx*4)/=vel_par%nz*vel_par%nx*4) call seperr("velocity read wrong4")
      do ix=1,nx
        dtmp_x=(ix-1)*vel_par%mig_dx-(ix_center-1)*vel_par%mig_dx
        do iz=1,nz
          dtmp_z=(iz-1)*vel_par%mig_dz
          tmp_x=min_x+dtmp_x*cos(theta)+dtmp_z*sin(theta)
          tmp_x=(tmp_x-tmp_min_x)/vel_par%dx+1.0
          tmp_ix=int(tmp_x);  tmp_wx=tmp_x-tmp_ix
          if (tmp_ix <=1) then
            tmp_ix=1;  tmp_wx=1.0
          end if
          if (tmp_ix >= tmp_vnx ) then 
            tmp_ix=tmp_vnx-1;  tmp_wx=0.5
          end if

          tmp_z=min_z-dtmp_x*sin(theta)+dtmp_z*cos(theta)
          tmp_z=(tmp_z-vel_par%min_z)/vel_par%dz+1.0
          tmp_iz=int(tmp_z); tmp_wz=tmp_z-tmp_iz
          if (tmp_iz < 1) then
            tmp_iz=1;   tmp_wz=0.5
          end if
          if (tmp_iz >= vel_par%nz ) then
            tmp_iz=vel_par%nz-1; tmp_wz=0.5
          end if
          vel(iz,ix,iy)=tmp_vel(tmp_iz,tmp_ix,1)*(1-tmp_wz)*(1-tmp_wx)+ &
                        tmp_vel(tmp_iz+1,tmp_ix,1)*tmp_wz*(1-tmp_wx)  + &
                        tmp_vel(tmp_iz,tmp_ix+1,1)*(1-tmp_wz)*tmp_wx  + &
                        tmp_vel(tmp_iz+1,tmp_ix+1,1)*tmp_wz*tmp_wx     
        end do
      end do
    end do
    deallocate(tmp_vel)
  else
    allocate(tmp_vel(vel_par%nz,vel_par%nx,1))
    tmp_min_x=vel_par%min_x
    tmp_vnx=vel_par%nx
  
    do iy=1,ny
      n_block=(iy-1)*vel_par%nx
      if (sseek_block("Velocity",n_block,vel_par%nz*4,0)/= n_block) call seperr("velocity seek wrong5")
      if(sreed("Velocity",tmp_vel(:,:,1),vel_par%nz*vel_par%nx*4)/=vel_par%nz*vel_par%nx*4) call seperr("velocity read wrong6")
      do ix=nx,1,-1
        dtmp_x=(nx-ix)*vel_par%mig_dx-(ix_center-1)*vel_par%mig_dx
        do iz=1,nz
          dtmp_z=(iz-1)*vel_par%mig_dz
          tmp_x=min_x-dtmp_x*cos(abs(theta))-dtmp_z*sin(abs(theta))
          tmp_x=(tmp_x-tmp_min_x)/vel_par%dx+1.0
          tmp_ix=int(tmp_x);  tmp_wx=tmp_x-tmp_ix
          if (tmp_ix <1) then
            tmp_ix=1;  tmp_wx=0.5
          end if
          if (tmp_ix >= tmp_vnx ) then 
            tmp_ix=tmp_vnx-1;  tmp_wx=0.5
          end if

          tmp_z=min_z-dtmp_x*sin(abs(theta))+dtmp_z*cos(abs(theta))
          tmp_z=(tmp_z-vel_par%min_z)/vel_par%dz+1.0
          tmp_iz=int(tmp_z); tmp_wz=tmp_z-tmp_iz
          if (tmp_iz < 1) then
            tmp_iz=1;   tmp_wz=0.5
          end if
          if (tmp_iz >= vel_par%nz ) then
            tmp_iz=vel_par%nz-1; tmp_wz=0.5
          end if
          vel(iz,nx-ix+1,iy)=tmp_vel(tmp_iz,tmp_ix,1)*(1-tmp_wz)*(1-tmp_wx)+ &
                        tmp_vel(tmp_iz+1,tmp_ix,1)*tmp_wz*(1-tmp_wx)  + &
                        tmp_vel(tmp_iz,tmp_ix+1,1)*(1-tmp_wz)*tmp_wx  + &
                        tmp_vel(tmp_iz+1,tmp_ix+1,1)*tmp_wz*tmp_wx       
        end do
      end do
    end do
    deallocate(tmp_vel)
  end if
end if
end subroutine velocity_read

subroutine velocity_mod_release
end subroutine velocity_mod_release

end module velocity_mod
